@@ -1,24 +1,14 @@
-// Copyright 2013 Google Inc. All Rights Reserved.
-//
-// Licensed under the Apache License, Version 2.0 (the "License");
-// you may not use this file except in compliance with the License.
-// You may obtain a copy of the License at
-//
-// http://www.apache.org/licenses/LICENSE-2.0
-//
-// Unless required by applicable law or agreed to in writing, software
-// distributed under the License is distributed on an "AS IS" BASIS,
-// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-// See the License for the specific language governing permissions and
-// limitations under the License.
-//
+/* Copyright 2013 Google Inc. All Rights Reserved.
+
+   Distributed under MIT license.
+   See file LICENSE for detail or copy at https://opensource.org/licenses/MIT
+*/
+
 // Function to find backward reference copies.
 
 #ifndef BROTLI_ENC_BACKWARD_REFERENCES_H_
 #define BROTLI_ENC_BACKWARD_REFERENCES_H_
 
-#include <vector>
-
 #include "./hash.h"
 #include "./command.h"
 #include "./types.h"
@@ -31,6 +21,7 @@ namespace brotli {
 // by this call.
 void CreateBackwardReferences(size_t num_bytes,
                               size_t position,
+                              bool is_last,
                               const uint8_t* ringbuffer,
                               size_t ringbuffer_mask,
                               const size_t max_backward_limit,
@@ -38,10 +29,10 @@ void CreateBackwardReferences(size_t num_bytes,
                               Hashers* hashers,
                               int hash_type,
                               int* dist_cache,
-                              int* last_insert_len,
+                              size_t* last_insert_len,
                               Command* commands,
                               size_t* num_commands,
-                              int* num_literals);
+                              size_t* num_literals);
 
 }  // namespace brotli
 
